/* Copyright (c) 2008 River Tarnell <river@wikimedia.org>. */
/*
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely. This software is provided 'as-is', without any express or implied
 * warranty.
 */
/* $Id: acceptor_thread.cc 52 2008-08-25 21:21:08Z river $ */

#include	<sys/types.h>
#include	<sys/socket.h>
#include	<sys/un.h>
#include	<netinet/in.h>

#include	<fcntl.h>
#include	<netdb.h>

#include	"acceptor_thread.h"
#include	"request_thread.h"
#include	"exception.h"
#include	"logger.h"

namespace {

extern "C" void *
start_acceptor(void *arg)
{
	acceptor_thread *thr = static_cast<acceptor_thread *>(arg);
	thr->start();
	return NULL;
}

} // anonymous namespace

acceptor_thread::acceptor_thread(
		std::string const &host,
		std::string const &service,
		boost::function<void (int, configp)> handler)
	: handler_(handler)
	, fd_(-1)
{
	int r;
	struct addrinfo hints, *res, *iter;
	std::memset(&hints, 0, sizeof(hints));
	hints.ai_flags = AI_PASSIVE | AI_ADDRCONFIG;
	hints.ai_socktype = SOCK_STREAM;

	if ((r = getaddrinfo(host.c_str(), service.c_str(),
			&hints, &res)) != 0) {
		throw gai_exception(str(boost::format("cannot resolve [%s]:%s")
				% host % service), r);
	}

	for (iter = res; iter; iter = iter->ai_next) {
		char ahost[NI_MAXHOST], aserv[NI_MAXSERV];
		int lsnsock;

		getnameinfo(iter->ai_addr, iter->ai_addrlen, 
				ahost, sizeof(ahost), aserv, sizeof(aserv),
				NI_NUMERICHOST | NI_NUMERICSERV);

		if ((lsnsock = socket(iter->ai_family, iter->ai_socktype, iter->ai_protocol)) == -1)
			throw errno_exception(
				str(boost::format("[%s]:%s: cannot create socket")
				% ahost % aserv));

		int one = 1;
		setsockopt(lsnsock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
		fcntl(lsnsock, F_SETFD, FD_CLOEXEC);

		if (bind(lsnsock, iter->ai_addr, iter->ai_addrlen) == -1)
			throw errno_exception(
				str(boost::format("[%s]:%s: cannot bind")
				% ahost % aserv));

		if (listen(lsnsock, 50) == -1)
			throw errno_exception(
				str(boost::format("[%s]:%s: cannot listen")
				% ahost % aserv));

		fd_ = lsnsock;
	}

	freeaddrinfo(res);
}

acceptor_thread::acceptor_thread(
		std::string const &path,
		boost::function<void (int, configp)> handler)
	: handler_(handler)
	, fd_(-1)
{
	struct sockaddr_un addr;
	int lsnsock;
	std::memset(&addr, 0, sizeof(addr));
	std::strncpy(addr.sun_path, path.c_str(), sizeof(addr.sun_path) - 1);
	addr.sun_family = AF_UNIX;

	unlink(addr.sun_path);

	if ((lsnsock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
		throw errno_exception(str(boost::format("%s: cannot create socket") % path));

	int one = 1;
	setsockopt(lsnsock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
	fcntl(lsnsock, F_SETFD, FD_CLOEXEC);

	if (bind(lsnsock, (sockaddr *) &addr, sizeof(addr)) == -1)
		throw errno_exception(str(boost::format("%s: cannot bind") % path));

	if (listen(lsnsock, 50) == -1)
		throw errno_exception(str(boost::format("%s cannot listen") % path));

	fd_ = lsnsock;
}

acceptor_thread::~acceptor_thread()
{
	if (fd_ != -1)
		close(fd_);
}

void
acceptor_thread::run()
{
	pthread_t tid;
	pthread_create(&tid, NULL, start_acceptor, this);
}

void
acceptor_thread::start()
{
	int newfd;
	struct sockaddr addr;
	socklen_t addrlen = sizeof(addr);

	while ((newfd = accept(fd_, &addr, &addrlen)) != -1) {
		fcntl(newfd, F_SETFD, FD_CLOEXEC);
		
		configp conf = get_current_config();
		handler_(newfd, conf);
	}

	logger.error(boost::format("accept failed: %s") % std::strerror(errno));
}
