/* Copyright (c) 2008 River Tarnell <river@wikimedia.org>. */
/*
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely. This software is provided 'as-is', without any express or implied
 * warranty.
 */
/* $Id: config.cc 62 2008-08-28 14:36:30Z river $ */

#include	<string>
#include	<fstream>
#include	<cerrno>
#include	<iostream>

#include	<pwd.h>

#include	<boost/format.hpp>
#include	<boost/tokenizer.hpp>
#include	<boost/assign/list_of.hpp>
#include	<boost/bind.hpp>
#include	<boost/lexical_cast.hpp>
using boost::format;

#include	"config.h"
#include	"logger.h"

config *curconf;

void
intrusive_ptr_add_ref(config *conf)
{
	sw_atomic_inc(&conf->refcnt);
}

void
intrusive_ptr_release(config *conf)
{
	sw_atomic_t newval = sw_atomic_dec_nv(&conf->refcnt);
	if (newval == 0)
		delete conf;
}

void
install_new_config(config *newconf)
{
	newconf->refcnt = 1;
	void *oldconf = sw_atomic_swap_ptr_ov((void **)&curconf, newconf);
	if (oldconf) {
		intrusive_ptr_release((config *) oldconf);
	}
}

configp
get_current_config()
{
	return configp(curconf);
}

config::config()
	: nfiles(0)
	, max_procs(0)
	, max_procs_per_user(0)
	, max_q_per_user(0)
	, php_timeout(30)
	, server_timeout(30)
	, max_request_size(1024 * 1024 * 5)
	, log_request_errors(false)
	, refcnt(0)
{
}

configuration_loader::configuration_loader()
{
}

bool
configuration_loader::load(std::string const &filename, config &newconf)
{
	std::string line;
	std::ifstream file(filename.c_str());

	lineno_ = 0;
	file_ = filename;

	if (!file) {
		logger.error( 
			format("cannot open configuration file \"%s\": %s")
			% filename % std::strerror(errno));
		return false;
	}

	while (std::getline(file, line)) {
		lineno_++;

		if (line.empty() || line[0] == '#')
			continue;

		if (!parse_config_line(line, newconf))
			return false;
	}

	return true;
}

std::map<std::string, configuration_loader::confline_t>
	configuration_loader::conflines =
	boost::assign::map_list_of
		("listen",		boost::bind(&configuration_loader::f_listen, _1, _2, _3, _4))
		("status-listen",	boost::bind(&configuration_loader::f_status_listen, _1, _2, _3, _4))
		("sockdir",		boost::bind(&configuration_loader::f_sockdir, _1, _2, _3, _4))
		("docroot",		boost::bind(&configuration_loader::f_docroot, _1, _2, _3, _4))
		("userdir",		boost::bind(&configuration_loader::f_userdir, _1, _2, _3, _4))
		("server-type",		boost::bind(&configuration_loader::f_server_type, _1, _2, _3, _4))
		("max-procs",		boost::bind(&configuration_loader::f_max_procs, _1, _2, _3, _4))
		("max-procs-per-user",	boost::bind(&configuration_loader::f_max_procs_per_user, _1, _2, _3, _4))
		("max-q-per-user",	boost::bind(&configuration_loader::f_max_q_per_user, _1, _2, _3, _4))
		("php-timeout",		boost::bind(&configuration_loader::f_php_timeout, _1, _2, _3, _4))
		("server-timeout",	boost::bind(&configuration_loader::f_server_timeout, _1, _2, _3, _4))
		("max-request-size",	boost::bind(&configuration_loader::f_max_request_size, _1, _2, _3, _4))
		("log-request-errors",	boost::bind(&configuration_loader::f_log_request_errors, _1, _2, _3, _4))
		("fcgi-extension",	boost::bind(&configuration_loader::f_fcgi_extn, _1, _2, _3, _4))
		("admin-contact",	boost::bind(&configuration_loader::f_admin_contact, _1, _2, _3, _4))
		("pid-file",		boost::bind(&configuration_loader::f_pid_file, _1, _2, _3, _4))
		("nfiles",		boost::bind(&configuration_loader::f_nfiles, _1, _2, _3, _4))
	;

bool
configuration_loader::parse_config_line(
		std::string const &line,
		config &newconf)
{
	boost::escaped_list_separator<char> sep('\\', ' ', '"');
	boost::tokenizer<boost::escaped_list_separator<char> > tok(line, sep);
	std::vector<std::string> fields;

	std::copy(tok.begin(), tok.end(), std::back_inserter(fields));

	if (fields.empty())
		return false;

	std::map<std::string, std::string> opts;
	/*
	 * The format of opts is [var1=val,var2=val] opt args....
	 */
	if (fields[0][0] == '[') {
		if (fields[0][fields[0].size() - 1] != ']') {
			logger.error(
				format("\"%s\", line %d: missing terminating ']' for options list")
					% file_ % lineno_);
			return false;
		}
		
		std::string optlist = fields[0].substr(1, fields[0].size() - 2);
		boost::escaped_list_separator<char> sep('\\', ',', '"');
		boost::tokenizer<boost::escaped_list_separator<char> > tok(optlist, sep);
		std::vector<std::string> ofields;
		std::copy(tok.begin(), tok.end(), std::back_inserter(ofields));
		
		for (int i = 0, end = ofields.size(); i < end; ++i) {
			std::string key(ofields[i]), value;
			std::string::size_type n;
			if ((n = key.find('=')) != std::string::npos) {
				value = key.substr(n + 1);
				key = key.substr(0, n);
			}
			
			opts[key] = value;
			std::cerr << "key [" << key << "] value [" << value << "]\n";
		}
		
		fields.erase(fields.begin());
	}
	
	std::map<std::string, confline_t>::iterator it =
		conflines.find(fields[0]);

	if ((it = conflines.find(fields[0])) == conflines.end()) {
		logger.error(
			format("\"%s\", line %d: unknown directive: \"%s\"")
			% file_ % lineno_ % fields[0]);
		return false;
	}
	
	it->second(this, fields, opts, newconf);

	return true;
}

bool
configuration_loader::f_listen(
		fieldlist_t & fields,
		confopts_t &,
		config &newconf)
{
	if (fields.size() < 2 ||
	    (fields[1][0] == '/' && fields.size() != 2) ||
	    (fields[1][0] != '/' && fields.size() != 3)) {
		logger.error(
			format("\"%s\", line %d: usage: listen <address> <port>\n")
			% file_ % lineno_);
		logger.error(
			format("\"%s\", line %d:        listen <pathname>\n")
			% file_ % lineno_);
		return false;
	}

	conf_listener lsnr;
	lsnr.host = fields[1];
	if (fields.size() == 3)
		lsnr.port = fields[2];
	newconf.listeners.push_back(lsnr);
	return true;
}

bool
configuration_loader::f_status_listen(
		fieldlist_t &fields,
		confopts_t &,
		config &newconf)
{
	if (fields.size() < 2 ||
	    (fields[1][0] == '/' && fields.size() != 2) ||
	    (fields[1][0] != '/' && fields.size() != 3)) {
		logger.error(
			format("\"%s\", line %d: usage: status-listen <address> <port>\n")
			% file_ % lineno_);
		logger.error(
			format("\"%s\", line %d:        status-listen <pathname>\n")
			% file_ % lineno_);
		return false;
	}

	conf_listener lsnr;
	lsnr.host = fields[1];
	if (fields.size() == 3)
		lsnr.port = fields[2];
	newconf.status_listeners.push_back(lsnr);
	return true;
}

bool
configuration_loader::f_sockdir(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	return _set_string("sockdir <directory>", fields, opts, &newconf.sockdir);
}

bool
configuration_loader::f_userdir(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	return _set_string("userdir <directory>", fields, opts, &newconf.userdir);
}

bool
configuration_loader::f_docroot(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	return _set_string("docroot <directory>", fields, opts, &newconf.docroot);
}

bool
configuration_loader::f_max_procs(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	return _set_int("max-procs <number>", fields, opts, &newconf.max_procs);
}

bool
configuration_loader::f_max_procs_per_user(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	if (fields.size() != 2) {
		logger.error(
			format("\"%s\", line %d: usage: max-procs-per-user <number>\n")
			% file_ % lineno_);
		return false;
	}

	try {
		if (!opts.empty()) {
			std::string user;
			for (confopts_t::iterator it = opts.begin(), end = opts.end();
					it != end; ++it)
			{
				if (it->first != "user") {
					logger.error(
						format("\"%s\", line %d: unknown option \"%s\"")
							% file_ % lineno_ % it->first);
					return false;
				}
				
				user = it->second;
			}
			
			if (user.empty())
				return false;
			
			struct passwd *pwd;
			if ((pwd = getpwnam(user.c_str())) == NULL) {
				logger.error(
					format("\"%s\", line %d: unknown user \"%s\": %s")
						% file_ % lineno_ % user % std::strerror(errno));
				return false;
			}
			
			newconf.per_user_max_procs[pwd->pw_uid] = boost::lexical_cast<int>(fields[1]);
		} else
			newconf.max_procs_per_user = boost::lexical_cast<int>(fields[1]);
		return true;
	} catch (boost::bad_lexical_cast &) {
		logger.error(
			format("\"%s\", line %d: usage: max-procs-per-user <number>\n")
			% file_ % lineno_);
		return false;
	}

//	return _set_int("max-procs-per-user <number>", fields, opts, &newconf.max_procs_per_user);
}

bool
configuration_loader::f_max_q_per_user(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	return _set_int("max-q-per-user <number>", fields, opts, &newconf.max_q_per_user);
}

bool
configuration_loader::f_server_type(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	if (fields.size() != 2) {
		logger.error(
			format("\"%s\", line %d: usage: server-type <sjs|sjs-sfn|apache|lighttpd|zws>\n")
			% file_ % lineno_);
		return false;
	}

	if (fields[1] == "sjs" || fields[1] == "lighttpd")
		newconf.servtype = serv_sjs;
	else if (fields[1] == "apache")
		newconf.servtype = serv_apache;
	else if (fields[1] == "sjs-sfn")
		newconf.servtype = serv_sjs_script_filename;
	else if (fields[1] == "zws")
		newconf.servtype = serv_zeus;
	else {
		logger.error(
			format("\"%s\", line %d: usage: server-type <sjs|sjs-sfn|apache|lighttpd|zws>\n")
			% file_ % lineno_);
		return false;
	}

	return true;
}

bool
configuration_loader::f_php_timeout(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	return _set_int("php-timeout <number>", fields, opts, &newconf.php_timeout);
}

bool
configuration_loader::f_server_timeout(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	return _set_int("server-timeout <number>", fields, opts, &newconf.server_timeout);
}

bool
configuration_loader::f_nfiles(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	return _set_int("nfiles <number>", fields, opts, &newconf.nfiles);
}

bool
configuration_loader::f_max_request_size(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	return _set_int("max-request-size <number>", fields, opts, &newconf.max_request_size);
}

bool
configuration_loader::f_fcgi_extn(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	if (fields.size() != 2) {
		logger.error(
			format("\"%s\", line %d: usage: fcgi-extn <string>\n")
			% file_ % lineno_);
		return false;
	}

	if (fields[1].find('.') != std::string::npos) {
		logger.error(
			format("\"%s\", line %d: fcgi-extn should not contain '.'")
			% file_ % lineno_);
		return false;
	}

	newconf.fcgi_extn = fields[1];
}

bool
configuration_loader::f_log_request_errors(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	if (fields.size() != 1) {
		logger.error(
			format("\"%s\", line %d: usage: log-request-errors\n")
			% file_ % lineno_);
		return false;
	}

	newconf.log_request_errors = true;
	return true;
}

bool
configuration_loader::f_admin_contact(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	return _set_string("admin-contact <address>", fields, opts, &newconf.admin_contact);
}

bool
configuration_loader::f_pid_file(
		fieldlist_t &fields,
		confopts_t &opts,
		config &newconf)
{
	return _set_string("pid-file <filename>", fields, opts, &newconf.pid_file);
}

bool
configuration_loader::_set_int(
		std::string const &help,
		fieldlist_t &fields,
		confopts_t &opts,
		int *where)
{
	if (fields.size() != 2) {
		logger.error(
			format("\"%s\", line %d: usage: %s\n")
			% file_ % lineno_ % help);
		return false;
	}

	try {
		*where = boost::lexical_cast<int>(fields[1]);
		return true;
	} catch (boost::bad_lexical_cast &) {
		logger.error(
			format("\"%s\", line %d: usage: %s\n")
			% file_ % lineno_ % help);
		return false;
	}
}

bool
configuration_loader::_set_string(
		std::string const &help,
		fieldlist_t &fields,
		confopts_t &opts,
		std::string *where)
{
	if (fields.size() != 2) {
		logger.error(
			format("\"%s\", line %d: usage: %s\n")
			% file_ % lineno_ % help);
		return false;
	}

	*where = fields[1];
	return true;
}
