/* Copyright (c) 2008 River Tarnell <river@wikimedia.org>. */
/*
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely. This software is provided 'as-is', without any express or implied
 * warranty.
 */
/* $Id: main.cc 52 2008-08-25 21:21:08Z river $ */

#include	<iostream>
#include	<cstdio>
#include	<cerrno>
#include	<cstring>
#include	<map>
#include	<string>
#include	<functional>
#include	<iterator>
#include	<csignal>
#include	<fstream>

#include	<sys/types.h>
#include	<sys/socket.h>
#include	<sys/un.h>
#include	<sys/wait.h>
#include	<sys/stat.h>
#include	<sys/resource.h>

#include	<netinet/in.h>
#include	<arpa/inet.h>

#include	<pthread.h>
#include	<pwd.h>
#include	<grp.h>
#include	<fcntl.h>
#include	<netdb.h>
#include	<cstdio>

#include	<boost/format.hpp>

#include	"fcgi.h"
#include	"request_thread.h"
#include	"config.h"
#include	"version.h"
#include	"process_factory.h"
#include	"acceptor_thread.h"
#include	"stats_thread.h"
#include	"setup.h"
#include	"logger.h"

std::string conf(CONFDIR "/switchboard.conf");
bool load_configuration();

extern "C" void *
exit_thread(void *)
{
	/*
	 * This should really do better cleanup, but this is okay
	 * for now.
	 */
	sigset_t sigs;
	sigemptyset(&sigs);
	sigaddset(&sigs, SIGINT);
	sigaddset(&sigs, SIGTERM);
	int sig;
	sigwait(&sigs, &sig);
	
	logger.info("exiting...");

	configp conf = get_current_config();
	std::remove(conf->pid_file.c_str());
	std::exit(0);
}

extern "C" void *
sighup_thread(void *)
{
	int sig;
	sigset_t sigs;
	sigemptyset(&sigs);
	sigaddset(&sigs, SIGHUP);
	
	while (sigwait(&sigs, &sig) == 0) {
		logger.info("SIGHUP received, reloading configuration");
		if (load_configuration())
			logger.info("new configuration installed");
		else
			logger.error("cannot load configuration");
	}
	
	logger.error(boost::format("sigwait() failed: %s")
			% std::strerror(errno));
	std::exit(1);
	return NULL;
}

/*
 * Load a new configuration and atomically swap it with the current
 * configuration.
 */
bool
load_configuration()
{
	configuration_loader loader;
	config *newconf = new config;
	if (!loader.load(conf, *newconf)) {
		logger.error("cannot load configuration");
		delete newconf;
		return false;
	}
    
	bool okay = true;
	if (newconf->sockdir.empty()) {
		logger.error("sockdir must be specified");
		okay = false;
	}

	if (newconf->userdir.empty()) {
		logger.error("userdir must be specified");
		okay = false;
	}

	if (newconf->docroot.empty()) {
		logger.error("docroot must be specified");
		okay = false;
	}

	if (newconf->servtype == serv_unknown) {
		logger.error("server type not set");
		okay = false;
	}
	
	if (!okay) {
		delete newconf;
		return false;
	}

	install_new_config(newconf);
	return true;
}

int
main(int argc, char **argv)
{
	int c, dflag = 0, Sflag = 0;

	while ((c = getopt(argc, argv, "df:DS")) != -1) {
		switch(c) {
		case 'd':
			dflag = 1;
			break;
		case 'S':
			Sflag = 1;
			break;
		case 'f':
			conf = optarg;
			break;
		case 'D':
			logger.set_severity(sev_debug);
			break;
		default:
			return 1;
		}
	}	

	if (Sflag)
		logger.set_target(log_syslog);

	logger.info("switchboard " SB_VERSION " starting up...");

	/*
	 * Block SIGHUP in all threads.  This is required so we can sigwait() for
	 * it in the config loader thread, and be guaranteed another thread won't
	 * receive it.  Also need SIGINT, SIGTERM for exit_thread.
	 */
	sigset_t sigs;
	sigemptyset(&sigs);
	sigaddset(&sigs, SIGHUP);
	sigaddset(&sigs, SIGINT);
	sigaddset(&sigs, SIGTERM);
	pthread_sigmask(SIG_BLOCK, &sigs, NULL);
	
	if (!load_configuration()) {
		return 1;
	}

	logger.info(boost::format("loaded configuration from \"%s\"") % conf);

	/* Don't hold on to the reference, we know the config isn't going away
	 * at this point.
	 */
	config *mainconf = get_current_config().get();
	
	if (mainconf->nfiles > 0) {
		struct rlimit lim;
		if (getrlimit(RLIMIT_NOFILE, &lim) == -1) {
			logger.warn(boost::format("getrlimit failed: %s")
					% std::strerror(errno));
			lim.rlim_cur = 0;
			lim.rlim_max = mainconf->nfiles;
		}
		lim.rlim_cur = mainconf->nfiles;
		if (lim.rlim_cur > lim.rlim_max)
			lim.rlim_max = lim.rlim_cur;
		if (setrlimit(RLIMIT_NOFILE, &lim) == -1)
			logger.warn(boost::format("cannot set open file limit to %d: %s")
				% mainconf->nfiles % std::strerror(errno));
	}
	
	if (!mainconf->pid_file.empty()) {
		std::ifstream f(mainconf->pid_file.c_str());
		if (f) {
			int pid;
			if (f >> pid) {
				if (kill(pid, 0) != -1 || errno != ESRCH) {
					logger.error(boost::format("switchboard is already running (pid %d)")
							% pid);
					return 1;
				}
			}
			f.close();
		}

		std::ofstream of(mainconf->pid_file.c_str());
		of << getpid() << '\n';
	}

	for (std::size_t i = 0; i < mainconf->listeners.size(); ++i) {
		try {
			if (mainconf->listeners[i].host[0] == '/') {
				acceptor_thread *acc = new acceptor_thread(
					mainconf->listeners[i].host,
					request_thread::accept_handler);
				acc->run();
			} else {
				acceptor_thread *acc = new acceptor_thread(
					mainconf->listeners[i].host,
					mainconf->listeners[i].port,
					request_thread::accept_handler);
				acc->run();
			}
		} catch (std::exception &e) {
			logger.error(e.what());
			logger.error("cannot continue");
			return 1;
		}
	}

	for (std::size_t i = 0; i < mainconf->status_listeners.size(); ++i) {
		try {
			if (mainconf->status_listeners[i].host[0] == '/') {
				acceptor_thread *acc = new acceptor_thread(
					mainconf->status_listeners[i].host,
					stats_thread::accept_handler);
				acc->run();
			} else {
				acceptor_thread *acc = new acceptor_thread(
					mainconf->status_listeners[i].host,
					mainconf->status_listeners[i].port,
					stats_thread::accept_handler);
				acc->run();
			}
		} catch (std::exception &e) {
			logger.error(e.what());
			logger.error("cannot continue");
			return 1;
		}
	}

	if (!getuid()) {
		struct passwd *pw;
		struct group *gr;

		if ((pw = getpwnam(SB_USER)) == NULL) {
			logger.error(
				boost::format("cannot setuid to \"%s\": user lookup failed: %s")
				% SB_USER % std::strerror(errno));
			return 1;
		}

		if ((gr = getgrnam(SB_GROUP)) == NULL) {
			logger.error(
				boost::format("cannot setgid to \"%s\": group lookup failed: %s")
				% SB_USER % std::strerror(errno));
			return 1;
		}

		if (!getgid()) {
			if (setgid(gr->gr_gid) == -1) {
				logger.error(
					boost::format("cannot setgid to \"%s\": %s")
					% SB_GROUP % std::strerror(errno));
				return 1;
			}
		}

		if (initgroups(SB_USER, gr->gr_gid) == -1) {
			logger.error(
					boost::format("cannot initialise supplemental group list: %s")
					% std::strerror(errno));
			return 1;
		}

		if (!getuid()) {
			if (setuid(pw->pw_uid) == -1) {
				logger.error(
					boost::format("cannot setuid to \"%s\": %s")
					% SB_USER % std::strerror(errno));
				return 1;
			}
		}

		 logger.info(boost::format("now running as %s:%s") % SB_USER % SB_GROUP);
	}

	/*
	 * Make sure our sockdir is valid, and has the right permissions.
	 */
dostat:
	struct stat sb;
	if (stat(mainconf->sockdir.c_str(), &sb) == -1) {
		if (errno == ENOENT) {
			if (mkdir(mainconf->sockdir.c_str(), 0750) == -1) {
				logger.error(
					boost::format("cannot create sockdir \"%s\": %s")
					% mainconf->sockdir % std::strerror(errno));
				return 1;
			}

			logger.info(boost::format("created sockdir \"%s\"") % mainconf->sockdir);
			goto dostat;
		} else {
			logger.error(
				boost::format("cannot stat sockdir \"%s"": %s")
				% mainconf->sockdir % std::strerror(errno));
			return 1;
		}
	}

	if (sb.st_uid != getuid()) {
		logger.error(
			boost::format("sockdir \"%s\" is not owned by my uid %d")
			% mainconf->sockdir % getuid());
		return 1;
	}

	if (sb.st_gid != getgid()) {
		logger.error(
			boost::format("sockdir \"%s\" is not owned by my gid %d")
			% mainconf->sockdir % getgid());
		return 1;
	}

	if ((sb.st_mode & S_IRWXO) != 0) {
		logger.error(
			boost::format("sockdir \"%s\" has wrong mode (access for others)")
			% mainconf->sockdir);
		return 1;
	}


	if (!dflag) {
		logger.info("detaching to background...");

		switch (fork()) {
		case -1:
			logger.error(boost::format("cannot fork: %s") % std::strerror(errno));
			return 1;
		case 0:
			break;
		default:
			return 0;
		}

		setsid();
		chdir("/");
		int fd;
		if ((fd = open("/dev/null", O_RDWR, 0)) != -1) {
			dup2(fd, STDIN_FILENO);
			dup2(fd, STDOUT_FILENO);
			dup2(fd, STDERR_FILENO);
			if (fd > STDERR_FILENO)
				close(fd);
		}
	}

	std::signal(SIGPIPE, SIG_IGN);
	std::signal(SIGCHLD, SIG_IGN);

	pthread_t sighup_t;
	pthread_attr_t hup_attrs;
	pthread_attr_init(&hup_attrs);
	pthread_attr_setdetachstate(&hup_attrs, PTHREAD_CREATE_DETACHED);
	if (pthread_create(&sighup_t, &hup_attrs, sighup_thread, NULL) == -1) {
		logger.error(boost::format("can't create SIGHUP thread: %s")
				% std::strerror(errno));
		return 1;
	}

	pthread_t exit_t;
	pthread_attr_t exit_attrs;
	pthread_attr_init(&exit_attrs);
	pthread_attr_setdetachstate(&exit_attrs, PTHREAD_CREATE_DETACHED);
	if (pthread_create(&exit_t, &exit_attrs, exit_thread, NULL) == -1) {
		logger.error(boost::format("can't create exit thread: %s")
				% std::strerror(errno));
		return 1;
	}
	
	process_factory::instance().cleanup_thread();
}
