/* Copyright (c) 2008 River Tarnell <river@wikimedia.org>. */
/*
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely. This software is provided 'as-is', without any express or implied
 * warranty.
 */
/* $Id: process_factory.cc 62 2008-08-28 14:36:30Z river $ */

#include	<sys/stat.h>
#include	<pwd.h>

#include	<cerrno>
#include	<iostream>
#include	<sstream>

#include	<boost/lambda/lambda.hpp>

#include	"process_factory.h"
#include	"config.h"
#include	"util.h"
#include	"setup.h"
#include	"swatomic.h"
#include	"stats_thread.h"

process_factory &
process_factory::instance()
{
	static process_factory inst;
	return inst;
}

process_factory::process_factory()
	: id_(0)
{
}

process_factory::~process_factory()
{
}

void
process_factory::cleanup_thread()
{
	for (;;) {
		sleep(30);

		lock l(lock_);

		std::time_t oldest = std::time(0) - 30;

		freelist_t::index<by_released>::type &idx = freelist_.get<by_released>();
		std::pair<
			freelist_t::index<by_released>::type::iterator,
			freelist_t::index<by_released>::type::iterator>
			range = idx.range(mi::unbounded, boost::lambda::_1 <= oldest);

		sw_atomic_add(&stats.nreaped, std::distance(range.first, range.second));
		idx.erase(range.first, range.second);
	}
}

processp
process_factory::get_process(
	std::map<std::string, std::string> &params,
	configp const &conf)
{
	std::map<std::string, std::string>::iterator it;

	uid_t uid;
	gid_t gid;

	std::string script_path;
	std::string program(PHP_BIN);

	/*
	 * Trying to find the script from path the env is a mess.  Under SJS 
	 * web server, we take the value from SCRIPT_NAME, and translate it 
	 * into a path on disk using the 'docroot' and 'userdir' configuration 
	 * options.  Under Apache, this doesn't work, because SCRIPT_NAME 
	 * contains garbage.  Instead we take PATH_TRANSLATED, which is the 
	 * on-disk path with the PATH_INFO appended, and remove path components 
	 * from it until we end up with a path which exists.
	 *
	 * Other web servers might require different handling; I haven't tested 
	 * any other than SJS and Apache.
	 */
	if (conf->servtype == serv_apache) {
		if ((it = params.find("PATH_TRANSLATED")) == params.end())
			throw creation_failure("PATH_TRANSLATED not specified");

		std::string s = it->second;
		struct stat sb;

		while (stat(s.c_str(), &sb) == -1
			&& errno == ENOTDIR)
		{
			std::string::size_type n;
			if ((n = s.rfind('/')) == std::string::npos)
				throw creation_failure("script not found");

			s.erase(n);
		}

		script_path = s;
	} else if (conf->servtype == serv_zeus) {
		if ((it = params.find("REDIRECT_PATH")) == params.end())
			throw creation_failure("REDIRECT_PATH not specified");
		script_path = it->second;
		params["SCRIPT_FILENAME"] = script_path;
		/* 
		 * REQUEST_URI is REDIRECT_URL with PATH_INFO appended.
		 * Strip this off, use it for PATH_INFO and append it to PATH_TRANSLATED
		 * to let PHP's script location logic work.
		 */
		std::string request_uri, redirect_url;
		if ((it = params.find("REQUEST_URI")) != params.end())
			request_uri = it->second;
		if ((it = params.find("REDIRECT_URL")) != params.end())
			redirect_url = it->second;
		if (!request_uri.empty() && !redirect_url.empty()) {
			if (request_uri.substr(0, redirect_url.size()) == redirect_url) {
				std::string path_info = request_uri.substr(redirect_url.size());
				params["PATH_INFO"] = path_info;
				params["PATH_TRANSLATED"] = script_path + path_info;
			}
		}
	} else if (conf->servtype == serv_sjs) {
		if ((it = params.find("SCRIPT_NAME")) == params.end())
			throw creation_failure("neither SCRIPT_NAME nor PATH_TRANSLATED specified");

		std::string script_name = it->second;
		if (script_name.empty())
			throw creation_failure("SCRIPT_NAME is empty");

		if (script_name.size() >= 2 &&
		    script_name[0] == '/' && script_name[1] == '~') {
			/*
			 * The format is /~user/path/to/script.php
			 * We need to change it to 
			 * /home/user/public_html/path/to/script.php.
			 */
			std::string username;
			std::string script;
			script_name.erase(script_name.begin(), script_name.begin() + 2);
			if (script_name.empty())
				throw creation_failure("invalid SCRIPT_NAME");
			std::string::size_type n = script_name.find('/');
			username.assign(script_name.begin(), script_name.begin() + n);
			script.assign(script_name.begin() + n + 1, script_name.end());

			struct passwd *pwd;
			if ((pwd = getpwnam(username.c_str())) == NULL)
				throw creation_failure("user does not exist");
			script_path = std::string(pwd->pw_dir) + '/' + conf->userdir
				+ '/' + script;
		} else {
			/*
			 * Script is relative to docroot.
			 */
			script_path = conf->docroot + script_name;
		}

		params["SCRIPT_FILENAME"] = script_path;
	} else if (conf->servtype == serv_sjs_script_filename) {
		if ((it = params.find("HTTP_SCRIPT_FILENAME")) == params.end())
			throw creation_failure("HTTP_SCRIPT_FILENAME not specified");

		script_path = it->second;
		params["SCRIPT_FILENAME"] = script_path;
	}

	/*
	 * If the script ends with fcgi_extn, assume it's a FastCGI program 
	 * itself, rather than a PHP script.
	 */
	if (!conf->fcgi_extn.empty()) {
		std::string::size_type n;
		if ((n = script_path.rfind('.')) != std::string::npos) {
			std::string extn = script_path.substr(n + 1);
			if (extn == conf->fcgi_extn) {
				program = script_path;
			}
		}
	}

	struct stat sb;
	if (lstat(script_path.c_str(), &sb) < 0)
		throw creation_failure("script doesn't exist");

	uid = sb.st_uid;
	gid = sb.st_gid;

	/*
	 * Make sure the path is under the docroot or the 
	 * user's userdir.
	 */
	std::string x = conf->docroot + '/';
	if (script_path.substr(0, x.size()) != x) {
		struct passwd *pwd = getpwuid(uid);
		if (pwd == NULL)
			throw creation_failure("script owner doesn't exist");
		x = std::string(pwd->pw_dir) + '/' + conf->userdir + '/';
		if (script_path.substr(0, x.size()) != x)
			throw creation_failure("script not under docroot or userdir");
	}

	std::ostringstream strm;
	strm << conf->sockdir << '/' << "php_" << uid << "_" << id_++;

	/*
	 * Look for a process in the freelist.
	 */
	std::map<uid_t, int>::const_iterator puit;
	int mppu = conf->max_procs_per_user;
	if ((puit = conf->per_user_max_procs.find(uid)) != 
			conf->per_user_max_procs.end())
		mppu = puit->second;
	int const mp = conf->max_procs;
	int const mqpu = conf->max_q_per_user;

	lock l(lock_);
	int &thisuser = peruser_[uid];
	int &thisq = peruser_waiting_[uid];

	if (mqpu > 0 && thisq > mqpu)
		throw creation_failure("too many queued processes for user");

	thisq++;

	while ( (mppu > 0 && thisuser >= mppu) ||
		(mp > 0 && curprocs_ > mp)) 
	{
		pthread_cond_wait(&curprocs_cond, &lock_);
	}

	peruser_[uid]++;
	curprocs_++;
	thisq--;

	freelist_t::index<by_uid>::type &idx = freelist_.get<by_uid>();
	freelist_t::index<by_uid>::type::iterator fit;
	if ((fit = idx.find(boost::make_tuple(uid, program))) != idx.end()) {
		processp p = fit->proc;
		idx.erase(fit);
		return p;
	}

	sw_atomic_inc(&stats.nnewprocs);
	return processp(new process(uid, gid, strm.str(), program));
}

void
process_factory::release_process(processp proc)
{
	free_process p;
	p.uid = proc->uid();
	p.released = std::time(0);
	p.proc = proc;
	p.program = proc->program();

	lock lck(lock_);
	freelist_.push_back(p);
	curprocs_--;
	peruser_[p.uid]--;
	pthread_cond_broadcast(&curprocs_cond);
}
	
void
process_factory::destroy_process(processp proc)
{
	uid_t uid = proc->uid();

	lock lck(lock_);
	curprocs_--;
	peruser_[uid]--;
	pthread_cond_broadcast(&curprocs_cond);
}
