/* Copyright (c) 2008 River Tarnell <river@wikimedia.org>. */
/*
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely. This software is provided 'as-is', without any express or implied
 * warranty.
 */
/* $Id: request_thread.cc 56 2008-08-26 18:45:22Z river $ */

#include	<iostream>
#include	<cerrno>

#include	<sys/types.h>
#include	<sys/socket.h>
#include	<sys/un.h>

#include	<fcntl.h>
#include	<pthread.h>

#include	<boost/format.hpp>

#include	"request_thread.h"
#include	"fcgi.h"
#include	"process_factory.h"
#include	"util.h"
#include	"config.h"
#include	"error.h"
#include	"exception.h"
#include	"stats_thread.h"
#include	"swatomic.h"
#include	"logger.h"

namespace {

extern "C" void*
do_start_thread(void *arg) {
	request_thread *req = static_cast<request_thread *>(arg);
	try {
		/*
		 * We try to handle several requests on this socket,
		 * if possible.
		 */
		for (;;)
			if (!req->start_request())
				break;
	} catch (std::exception &e) {
		if (req->conf()->log_request_errors) {
			logger.error(boost::format("request failed: %s") % e.what());
		}
		sw_atomic_inc(&stats.nfailed);
	}

	sw_atomic_inc(&stats.nrequests);
	delete req;

	return NULL;
}

} // anonymous namespace

void
request_thread::accept_handler(int fd, configp const &conf)
{
	request_thread *req = new request_thread(fd, conf);
	req->start();
}

request_thread::request_thread(int fd, configp const &conf)
	: fd_(fd)
	, rid_(-1)
	, cfd_(-1)
	, conf_(conf)
{
}

request_thread::~request_thread()
{
	if (cfd_ != -1)
		close(cfd_);

	if (fd_ != -1)
		close(fd_);
}

void
request_thread::start()
{
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	pthread_create(&tid_, &attr, &do_start_thread, static_cast<void *>(this));
}

bool
request_thread::start_request()
{
	rid_ = -1;
	cfd_ = -1;
	stdin_.clear();
	paramdata_.clear();
	params_.clear();
	
	fcgi::record rec;

	/*
	 * Start by parsing the request header.
	 */
	errno = 0;
	if (!fcgi::read_fcgi_record(fd_, &rec, conf_->server_timeout)) {
		/*
		 * errno = 0 means the connection was closed.  This isn't an
		 * error, so we don't throw for it.
		 */
		if (errno == 0 || errno == ETIMEDOUT)
			return false;
		throw errno_exception("error reading initial record");
	}

	if (rec.version() != 1)
		throw request_exception("unknown FCGI version");

	rid_ = rec.request_id();

	switch (rec.type()) {
	case fcgi::rectype::begin_request:
		try {
			handle_normal_request(rec);
		} catch (std::exception &e) {
			if (conf_->log_request_errors) {
				logger.error(boost::format("request failed: %s") % e.what());
			}
			sw_atomic_inc(&stats.nfailed);

			fcgi::record r;

			r.version_ = 1;
			r.paddingLength_ = 0;
			r.reserved_ = 0;
			r.type_ = fcgi::rectype::stdout_;
			r.request_id(rid_);

			error errfmt(params_, conf_);
			std::string err =
"Status: HTTP/1.1 500 Internal server error\r\n"
"Content-Type: text/html\r\n"
"\r\n"
			+ errfmt.format(e.what());

			std::copy(err.begin(), err.end(), std::back_inserter(r.contentData));
			r.content_length(r.contentData.size());
			if (!fcgi::write_fcgi_record(fd_, r, conf_->server_timeout))
				return false;

			r.contentData.clear();
			r.content_length(0);
			if (!fcgi::write_fcgi_record(fd_, r, conf_->server_timeout))
				return false;

			r.type_ = fcgi::rectype::end_request;
			r.contentData.resize(8);
			r.contentData[0] = 0;
			r.contentData[1] = 0;
			r.contentData[2] = 0;
			r.contentData[3] = 0;
			r.contentData[4] = 0;
			r.content_length(8);

			if (!fcgi::write_fcgi_record(fd_, r, conf_->server_timeout))
				return false;
			
			return false;
		}

		return true;

	case fcgi::rectype::get_values:
		handle_get_values(rec);
		return true;

	default:
		throw request_exception("unexpected request type at this point");
	}
}

void
request_thread::handle_normal_request(fcgi::record &initial)
{
	fcgi::record rec;
	fcgi::begin_request_payload *pl = reinterpret_cast<fcgi::begin_request_payload *>(&initial.contentData[0]);

	if (initial.content_length() != 8)
		throw request_exception("begin request had unexpected content length");

	switch (pl->role()) {
	case fcgi::role::responder:
		break;
	
	default:
		throw request_exception("begin_request had unexpected role");
	}

	std::size_t reqsize = 0;

	/*
	 * Now we should receive at least one params.
	 */
	for (;;) {
		if (!fcgi::read_fcgi_record(fd_, &rec, conf_->server_timeout))
			throw errno_exception("error reading params from server");

		if (rec.type() != fcgi::rectype::params)
			throw request_exception("unexpected request type at this point");

		if (rec.content_length() == 0)
			break;

		if (reqsize + rec.content_length() > conf_->max_request_size)
			throw request_exception("request too big");
		reqsize += rec.content_length();

		paramdata_.insert(paramdata_.end(), rec.contentData.begin(), rec.contentData.end());
	}

	/*
	 * And at least one stdin.
	 */
	for (;;) {
		if (!fcgi::read_fcgi_record(fd_, &rec, conf_->server_timeout))
			throw errno_exception("error reading params from server");

		if (rec.type() != fcgi::rectype::stdin_)
			throw request_exception("unexpected request type at this point");

		if (rec.content_length() == 0)
			break;

		if (reqsize + rec.content_length() > conf_->max_request_size)
			throw request_exception("request too big");
		reqsize += rec.content_length();

		stdin_.insert(stdin_.end(), rec.contentData.begin(), rec.contentData.end());
	}

	/*
	 * Handle the actual request.
	 */

	int tries = 0;
	static int const max_tries = 10;

	do {
		try {
			handle_normal_request_child();

			close(cfd_);
			cfd_ = -1;
			return;
		} catch (errno_exception &e) {
			close(cfd_);
			cfd_ = -1;
			if (tries++ >= max_tries || e.err() == ETIMEDOUT)
				throw;
		} catch (std::exception &e) {
			close(cfd_);
			cfd_ = -1;
			if (tries++ >= max_tries)
				throw;
		}
	} while (true);
}

void
request_thread::handle_normal_request_child()
{
	fcgi::decode_params(paramdata_.begin(), paramdata_.end(), std::inserter(params_, params_.begin()));

#if 0
	for (std::map<std::string, std::string>::iterator it = params_.begin(), end = params_.end(); it != end; ++it)
		std::cout << boost::format("[%s] = [%s]\n") % it->first % it->second;
#endif
	
	process_ = process_factory::instance().get_process(params_, conf_);
	process_releaser p(process_);

	if ((cfd_ = process_->connect()) == -1)
		throw errno_exception("can't connect to FastCGI process");

	/*
	 * Write the request to the child.
	 */
	fcgi::record r_begin;
	fcgi::record r_params;
	fcgi::record r_stdin;

	r_begin.version_ = r_params.version_ = r_stdin.version_ = 1;
	r_begin.paddingLength_ = r_params.paddingLength_ = r_stdin.paddingLength_ = 0;
	r_begin.reserved_ = r_params.reserved_ = r_stdin.reserved_ = 0;

	r_begin.request_id(rid_);
	r_begin.type_ = fcgi::rectype::begin_request;
	r_begin.content_length(8);
	r_begin.contentData.resize(8);
	r_begin.contentData[0] = 0;
	r_begin.contentData[1] = 1;
	r_begin.contentData[2] = 0;
	if (!fcgi::write_fcgi_record(cfd_, r_begin, conf_->php_timeout))
		throw errno_exception("error writing begin request to child");

	std::vector<unsigned char> newparams;
	std::vector<unsigned char>::iterator dit, dend;
	fcgi::encode_params(params_.begin(), params_.end(), std::back_inserter(newparams));
	r_params.request_id(rid_);
	r_params.type_ = fcgi::rectype::params;

	dit = newparams.begin();
	dend = newparams.end();

	while (dit != dend) {
		int now = std::min(65535, (int) std::distance(dit, dend));
		r_params.contentData.resize(now);
		std::copy(dit, dit + now, r_params.contentData.begin());
		r_params.content_length(now);
		if (!fcgi::write_fcgi_record(cfd_, r_params, conf_->php_timeout))
			throw errno_exception("error writing params to child");

		dit += now;
	}

	r_params.contentData.clear();
	r_params.content_length(0);
	if (!fcgi::write_fcgi_record(cfd_, r_params, conf_->php_timeout))
		throw errno_exception("error writing end of params to child");

	
	r_stdin.request_id(rid_);
	r_stdin.type_ = fcgi::rectype::stdin_;

	dit = stdin_.begin();
	dend = stdin_.end();

	/*
	 * Only 64K in allowed in one record.
	 */
	while (dit != dend) {
		int now = std::min(65535, (int) std::distance(dit, dend));
		r_stdin.contentData.resize(now);
		std::copy(dit, dit + now, r_stdin.contentData.begin());
		r_stdin.content_length(now);
		if (!fcgi::write_fcgi_record(cfd_, r_stdin, conf_->php_timeout))
			throw errno_exception("error writing stdin to child");

		dit += now;
	}

	r_stdin.contentData.clear();
	r_stdin.content_length(0);
	if (!fcgi::write_fcgi_record(cfd_, r_stdin, conf_->php_timeout))
		throw errno_exception("error writing stdin to child");

	if (!r_stdin.contentData.empty()) {
		r_stdin.contentData.clear();
		r_stdin.content_length(0);
		if (!fcgi::write_fcgi_record(cfd_, r_stdin, conf_->php_timeout))
			throw errno_exception("error writing end of stdin to child");
	}

	/*
	 * Now read response requests and forward them to the server.
	 */
	fcgi::record resp;
	for (;;) {
		if (!fcgi::read_fcgi_record(cfd_, &resp, conf_->php_timeout))
			throw errno_exception("couldn't read record from child");

		if (resp.request_id() != rid_)
			throw request_exception("child sent incorrect request id");

		if (!fcgi::write_fcgi_record(fd_, resp, conf_->server_timeout))
			throw errno_exception("couldn't write record to server");

		if (resp.type() == fcgi::rectype::end_request)
			break;
	}

	p.release();
	process_.reset();
}

void
request_thread::handle_get_values(fcgi::record &initial)
{
	std::map<std::string, std::string> request;
	std::map<std::string, std::string> response;

	fcgi::decode_params(initial.contentData.begin(), initial.contentData.end(),
			std::inserter(request, request.begin()));

	for (std::map<std::string, std::string>::iterator
			it = request.begin(), end = request.end();
			it != end; ++it)
	{
		if (it->first == "FCGI_MAX_CONNS")
			response["FCGI_MAX_CONNS"] = "100";
		else if (it->first == "FCGI_MAX_REQS")
			response["FCGI_MAX_REQS"] = "100";
		else if (it->first == "FCGI_MPXS_CONNS")
			response["FCGI_MPXS_CONNS"] = "0";
	}

	fcgi::record rec;
	std::vector<unsigned char> responsedata;
	fcgi::encode_params(response.begin(), response.end(),
			std::back_inserter(rec.contentData));
	rec.version_ = 1;
	rec.type_ = fcgi::rectype::get_values_result;
	rec.requestId0_ = initial.requestId0_;
	rec.requestId1_ = initial.requestId1_;
	rec.reserved_ = 0;
	rec.content_length(rec.contentData.size());
	fcgi::write_fcgi_record(fd_, rec, conf_->server_timeout);
}
