/* Copyright 2008 The Apache Software Foundation. */
/* Copyright 2008 River Tarnell <river@wikimedia.org */
/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * suexec.c -- "Wrapper" support program for suEXEC behaviour for Apache
 */

#define SB_SUEXEC_UMASK 022

#include	<sys/param.h>
#include	<sys/stat.h>
#include	<sys/types.h>
#include	<string.h>
#include	<time.h>
#include	<unistd.h>
#include	<stdio.h>
#include	<stdarg.h>
#include	<stdlib.h>
#include	<errno.h>

#include	"setup.h"

#ifdef HAVE_PROJECTS
# include	<sys/task.h>
# include	<project.h>
#endif

#ifdef HAVE_PAM
# include	<security/pam_appl.h>
#endif

#include	<pwd.h>
#include	<grp.h>

#if defined(PATH_MAX)
#define SB_MAXPATH PATH_MAX
#elif defined(MAXPATHLEN)
#define SB_MAXPATH MAXPATHLEN
#else
#define SB_MAXPATH 8192
#endif

#define SB_ENVBUF 256

extern char **environ;
static FILE *log = NULL;

static const char *const safe_env_lst[] =
{
    "TZ=",
    NULL
};


static void
err_output(int is_error, const char *fmt, va_list ap)
{
    time_t timevar;
    struct tm *lt;

    if (!log) {
        if ((log = fopen(SB_LOG_EXEC, "a")) == NULL) {
            fprintf(stderr, "suexec failure: could not open log file\n");
            perror("fopen");
            exit(1);
        }
    }

    if (is_error) {
        fprintf(stderr, "suexec policy violation: see suexec log for more "
                        "details\n");
    }

    time(&timevar);
    lt = localtime(&timevar);

    fprintf(log, "[%d-%.2d-%.2d %.2d:%.2d:%.2d]: ",
            lt->tm_year + 1900, lt->tm_mon + 1, lt->tm_mday,
            lt->tm_hour, lt->tm_min, lt->tm_sec);

    vfprintf(log, fmt, ap);

    fflush(log);
}

static void
log_err(const char *fmt,...)
{
    va_list ap;

    va_start(ap, fmt);
    err_output(1, fmt, ap); /* 1 == is_error */
    va_end(ap);
}

static void
log_no_err(const char *fmt,...)
{
    va_list ap;

    va_start(ap, fmt);
    err_output(0, fmt, ap); /* 0 == !is_error */
    va_end(ap);
}

#ifdef HAVE_PAM
static int
pam_null_conv(int num_msg, struct pam_message **msgs,
		struct pam_response **resp, void *my_data)
{
	return PAM_CONV_ERR;
}
#endif
	
static void
clean_env(void)
{
	char envbuf[512];
	char **cleanenv;
	char **ep;
	int cidx = 0;
	int idx;

	/* While cleaning the environment, the environment should be clean.
	 * (e.g. malloc() may get the name of a file for writing debugging info.
	 * Bad news if MALLOC_DEBUG_FILE is set to /etc/passwd.  Sprintf() may be
	 * susceptible to bad locale settings....)
	 * (from PR 2790)
	 */
	char **envp = environ;
	char *empty_ptr = NULL;

	environ = &empty_ptr; /* VERY safe environment */

	if ((cleanenv = (char **) calloc(SB_ENVBUF, sizeof(char *))) == NULL) {
		log_err("failed to malloc memory for environment\n");
		exit(120);
	}

	sprintf(envbuf, "PATH=%s", SB_SAFE_PATH);
	cleanenv[cidx] = strdup(envbuf);
	cidx++;
	sprintf(envbuf, "PHP_FCGI_MAX_REQUESTS=5000");
	cleanenv[cidx] = strdup(envbuf);
	cidx++;

	for (ep = envp; *ep && cidx < SB_ENVBUF-1; ep++) {
		for (idx = 0; safe_env_lst[idx]; idx++) {
			if (!strncmp(*ep, safe_env_lst[idx],
					strlen(safe_env_lst[idx]))) {
				cleanenv[cidx] = *ep;
				cidx++;
				break;
			}
		}
	}

	cleanenv[cidx] = NULL;

	environ = cleanenv;
}

int
main(int argc, char *argv[])
{
	uid_t uid;              /* user information          */
	gid_t gid;              /* target group placeholder  */
	char *prog;		/* program to start          */
	char *target_uname;     /* target user name          */
	char *target_gname;     /* target group name         */
	char *target_homedir;   /* target home directory     */
	char *actual_uname;     /* actual user name          */
	char *actual_gname;     /* actual group name         */
	struct passwd *pw;      /* password entry holder     */
	struct group *gr;       /* group entry holder        */
#ifdef HAVE_PROJECTS
	struct project proj;
	char nssbuf[PROJECT_BUFSZ];
#endif	/* HAVE_PROJECTS */
#ifdef HAVE_PAM
	int err;
	pam_handle_t *pamh;
	struct pam_conv conv = { pam_null_conv, NULL };
#endif	/* HAVE_PAM */
	/*
	 * Start with a "clean" environment
	 */
	clean_env();

	/*
	 * Check existence/validity of the UID of the user
	 * running this program.  Error out if invalid.
	 */
	uid = getuid();
	if ((pw = getpwuid(uid)) == NULL) {
		log_err("crit: invalid uid: (%ld)\n", (long) uid);
		exit(102);
	}

	/*
	 * If there are a proper number of arguments, set
	 * all of them to variables.  Otherwise, error out.
	 */
	if (argc < 4) {
		log_err("too few arguments\n");
		exit(101);
	}
	target_uname = argv[1];
	target_gname = argv[2];
	prog = argv[3];

	/*
	 * Check to see if the user running this program
	 * is the user allowed to do so as defined in
	 * suexec.h.  If not the allowed user, error out.
	 */
	if (strcmp(SB_USER, pw->pw_name)) {
		log_err("user mismatch (%s instead of %s)\n", pw->pw_name, SB_USER);
		exit(103);
	}

	/*
	 * Error out if the target username is invalid.
	 */
	if (strspn(target_uname, "1234567890") != strlen(target_uname)) {
		if ((pw = getpwnam(target_uname)) == NULL) {
			log_err("invalid target user name: (%s)\n", target_uname);
			exit(105);
		}
	} else {
		if ((pw = getpwuid(atoi(target_uname))) == NULL) {
			log_err("invalid target user id: (%s)\n", target_uname);
			exit(121);
		}
	}

	/*
	 * Error out if the target group name is invalid.
	 */
	if (strspn(target_gname, "1234567890") != strlen(target_gname)) {
		if ((gr = getgrnam(target_gname)) == NULL) {
			log_err("invalid target group name: (%s)\n", target_gname);
			exit(106);
		}
	} else {
		if ((gr = getgrgid(atoi(target_gname))) == NULL) {
			log_err("invalid target group id: (%s)\n", target_gname);
			exit(106);
		}
	}
	gid = gr->gr_gid;
	actual_gname = strdup(gr->gr_name);

	/*
	 * Save these for later since initgroups will hose the struct
	 */
	uid = pw->pw_uid;
	actual_uname = strdup(pw->pw_name);
	target_homedir = strdup(pw->pw_dir);

	/*
	 * Log the transaction here to be sure we have an open log
	 * before we setuid().
	 */
	log_no_err("uid: (%s/%s) gid: (%s/%s) cmd: %s\n",
		target_uname, actual_uname,
		target_gname, actual_gname,
		prog);

	/*
	 * Error out if attempt is made to execute as root or as
	 * a UID less than SB_UID_MIN.  Tsk tsk.
	 */
	if ((uid == 0) || (uid < SB_UID_MIN)) {
		log_err("cannot run as forbidden uid (%d/%s)\n", uid, prog);
		exit(107);
	}

	/*
	 * Error out if attempt is made to execute as root group
	 * or as a GID less than SB_GID_MIN.  Tsk tsk.
	 */
	if ((gid == 0) || (gid < SB_GID_MIN)) {
		log_err("cannot run as forbidden gid (%d/%s)\n", gid, prog);
		exit(108);
	}

#ifdef HAVE_PAM
	if ((err = pam_start("swexec", actual_uname, &conv, &pamh)) != PAM_SUCCESS) {
		log_err("failed to initialise PAM (%s: %s)\n",
				actual_uname, pam_strerror(pamh, err));
		exit(111);
	}
	
	if ((err = pam_acct_mgmt(pamh, PAM_SILENT)) != PAM_SUCCESS) {
		log_err("failed to start user session (%s: %s)\n",
			actual_uname, pam_strerror(pamh, err));
		exit(113);
	}
	
	if ((err = pam_setcred(pamh, PAM_ESTABLISH_CRED | PAM_SILENT)) != PAM_SUCCESS) {
		log_err("failed to set PAM credentials (%s: %s)\n",
				actual_uname, pam_strerror(pamh, err));
		exit(114);
	}
#endif	/* !HAVE_PAM */

#ifdef HAVE_PROJECTS
	if (getdefaultproj(actual_uname, &proj, nssbuf, sizeof(nssbuf)) == NULL) {
		log_err("failed to get default project (%s: %s)",
				actual_uname, strerror(errno));
		exit(113);
	}

	if (setproject(proj.pj_name, actual_uname, TASK_FINAL) != 0) {
		log_err("failed to set project (%s: %s)",
				actual_uname, strerror(errno));
		exit(114);
	}
#endif
		
	/*
	 * Change UID/GID here so that the following tests work over NFS.
	 *
	 * Initialize the group access list for the target user,
	 * and setgid() to the target group. If unsuccessful, error out.
	 */
	if (((setgid(gid)) != 0) || (initgroups(actual_uname, gid) != 0)) {
		log_err("failed to setgid (%ld: %s)\n", (long) gid, prog);
		exit(109);
	}

	/*
	 * setuid() to the target user.  Error out on fail.
	 */
	if ((setuid(uid)) != 0) {
		log_err("failed to setuid (%ld: %s)\n", (long) uid, prog);
		exit(110);
	}

	if (chdir(target_homedir) != 0) {
		log_err("cannot chdir to home directory (%s)\n", target_homedir);
		exit(112);
	}

#ifdef SB_SUEXEC_UMASK
	/*
	 * umask() uses inverse logic; bits are CLEAR for allowed access.
	 */
	if ((~SB_SUEXEC_UMASK) & 0022) {
		log_err("notice: SB_SUEXEC_UMASK of %03o allows "
			"write permission to group and/or other\n", SB_SUEXEC_UMASK);
	}
	umask(SB_SUEXEC_UMASK);
#endif /* SB_SUEXEC_UMASK */

	/*
	 * Be sure to close the log file so the CGI can't
	 * mess with it.  If the exec fails, it will be reopened
	 * automatically when log_err is called.  Note that the log
	 * might not actually be open if SB_LOG_EXEC isn't defined.
	 * However, the "log" call isn't ifdef'd so let's be defensive
	 * and assume someone might have done something with it
	 * outside an ifdef'd SB_LOG_EXEC block.
	 */
	if (log != NULL) {
		fclose(log);
		log = NULL;
	}

	/*
	 * Execute the command, replacing our image with its own.
	 */
	execl(prog, prog, NULL);

#ifdef HAVE_PAM
	pam_end(pamh, 0);
#endif
	
	/*
	 * (I can't help myself...sorry.)
	 *
	 * Uh oh.  Still here.  Where's the kaboom?  There was supposed to be an
	 * EARTH-shattering kaboom!
	 *
	 * Oh well, log the failure and error out.
	 */
	log_err("(%d)%s: exec failed (%s)\n", errno, strerror(errno), prog);
	exit(255);
}
