/* Copyright (c) 2008 River Tarnell <river@wikimedia.org>. */
/*
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely. This software is provided 'as-is', without any express or implied
 * warranty.
 */
/* $Id: switchstats.cc 42 2008-08-21 04:36:01Z river $ */

#include	<sys/types.h>
#include	<sys/socket.h>
#include	<sys/un.h>

#include	<unistd.h>
#include	<netdb.h>

#include	<iostream>
#include	<cerrno>
#include	<sstream>
#include	<cstring>

#include	<boost/format.hpp>
#include	<boost/archive/text_iarchive.hpp>

#include	"stats_thread.h"

namespace {

int
connect_unix(std::string const &path)
{
	struct sockaddr_un addr;
	int sock;
	std::memset(&addr, 0, sizeof(addr));
	std::strncpy(addr.sun_path, path.c_str(), sizeof(addr.sun_path) - 1);
	addr.sun_family = AF_UNIX;

	if ((sock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
		perror("socket");
		return -1;
	}

	if (connect(sock, (sockaddr *) &addr, sizeof(addr)) == -1) {
		std::cerr << boost::format("%s: connect: %s\n")
			% path % std::strerror(errno);
		return -1;
	}

	return sock;
}

int
connect_inet(std::string const &host, std::string const &service)
{
	int r, fd = -1;
	struct addrinfo hints, *res, *iter;
	std::memset(&hints, 0, sizeof(hints));
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_ADDRCONFIG;

	if ((r = getaddrinfo(host.c_str(), service.c_str(),
			&hints, &res)) != 0) {
		std::cerr << boost::format("cannot resolve [%s]:%s: %s\n")
				% host % service % gai_strerror(r);
		return -1;
	}

	for (iter = res; iter; iter = iter->ai_next) {
		char ahost[NI_MAXHOST], aserv[NI_MAXSERV];

		getnameinfo(iter->ai_addr, iter->ai_addrlen, 
				ahost, sizeof(ahost), aserv, sizeof(aserv),
				NI_NUMERICHOST | NI_NUMERICSERV);

		if ((fd = socket(iter->ai_family, iter->ai_socktype, iter->ai_protocol)) == -1) {
			std::cerr << boost::format("[%s]:%s: cannot create socket: %s\n")
				% ahost % aserv % std::strerror(errno);
			continue;
		}

		if (connect(fd, iter->ai_addr, iter->ai_addrlen) == -1) {
			std::cerr << boost::format("[%s]:%s: %s\n")
				% ahost % aserv % std::strerror(errno);
			close(fd);
			fd = -1;
			continue;
		}

		break;
	}

	freeaddrinfo(res);
	return fd;
}

} // anonymous namespace

int
main(int argc, char **argv)
{
	if (argc != 2) {
		std::cerr << boost::format("usage: %s <host:port | path>\n")
			% argv[0];
		return 1;
	}

	int fd;

	if (argv[1][0] == '/') {
		if ((fd = connect_unix(argv[1])) == -1)
			return 1;
	} else {
		std::string arg(argv[1]), host, port;
		std::string::size_type n;
		if ((n = arg.find(':')) == std::string::npos) {
			std::cerr << boost::format("usage: %s <host:port | path>\n") % argv[0];
			return 1;
		}

		host = arg.substr(0, n);
		port = arg.substr(n + 1);
		if ((fd = connect_inet(host, port)) == -1) {
			std::cerr << "cannot connect to any addresses\n";
			return 1;
		}
	}

	char buf[8192];
	std::vector<char> data;
	int i;

	while ((i = read(fd, buf, sizeof(buf))) > 0)
		std::copy(buf, buf + i, std::back_inserter(data));

	if (i == -1) {
		perror("read error");
		return 1;
	}

	std::stringstream strm;
	strm.write(&data[0], data.size());
	boost::archive::text_iarchive ia(strm);
	stats_t stats;
	ia >> stats;

	int days = 0, hours = 0, minutes = 0, seconds = 0;
	int uptime = std::time(0) - stats.startup, n = uptime;
	
	days = n / (60 * 60 * 24);
	n %= (60 * 60 * 24);
	hours = n / (60 * 60);
	n %= (60 * 60);
	minutes = n / 60;
	n %= 60;
	seconds = n;

	std::cout << boost::format("switchboard %s on %-20s Up %d+%02d:%02d:%02d\n")
		% stats.version % stats.hostname
		% days % hours % minutes % seconds;

	std::cout << boost::format("     requests:    %d (%.02f/sec.)\n")
		% stats.nrequests % (stats.nrequests / uptime);
	std::cout << boost::format("       failed:    %d (%.02f/sec.)\n")
		% stats.nfailed % (stats.nfailed / uptime);
	std::cout << boost::format("new processes:    %d (%.02f/sec.)\n")
		% stats.nnewprocs % (stats.nnewprocs / uptime);
	if (stats.nnewprocs && stats.nrequests)
		std::cout << boost::format("                  (%.02f%% efficiency)\n")
			% (100 - ((stats.nnewprocs / (double) stats.nrequests) * 100));
	std::cout << boost::format("processes reaped: %d\n") % stats.nreaped;
}
